from dyndis.multidispatch import MultiDispatch, RawNotImplemented, RawReturnValue
from dyndis.candidate import Self
from dyndis.util import AmbiguityError
from dyndis_data import __version__, __author__, __url__
